# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for OpenFOAM cases.

This module contains classes to load an OpenFOAM case into an object. The
OpenFOAM case object uses the functions provided via the foam module to
operate in the case, e.g., read and write data
"""
from sys import exit
from subprocess import run, CalledProcessError
import logging
from os import path
from typing import NoReturn, Iterable, Any, Tuple, Sequence, Dict, List
from numpy import ndarray
from jinja2 import Environment, FileSystemLoader
import yaml
from . import foam

# Enable logging for this module
log = logging.getLogger(__name__)


class FoamBaseCase:
    """A simple simulation case.

    This class contains parameters and functions for a plain simulation case.
    """

    def __init__(self, dname: str) -> NoReturn:
        """Initialize plain simulation case.

        Initializes a plain simulation case, check for validity of the case and
        store some basic parameters

        Args:
            dname: directory containing the OpenFOAM case.

        Parameters:
            _name (str): Name of the OpenFOAM case
            _path (str): Directory of the OpenFOAM case

        Returns:
            Nothing.
        """
        self._name = path.basename(path.abspath(dname).rstrip("/"))
        self._path = path.abspath(dname)

        if not foam.is_case(self._path):
            log.error(f"This is not a valid case in {self._path}.")
            exit(1)
        else:
            log.info(f"Created base case for {self._path}.")

    def name(self) -> str:
        """Return the name of the simulation case.

        Returns:
            Name of the simulation case.
        """
        return self._name

    def path(self) -> str:
        """Return the directory of the simulation case.

        Returns:
            Absolute path of directory of the simulation case.
        """
        return self._path

    def get_dictionary(self, fname) -> dict:
        """Return the dictionary in the file as python dictionary.

        Args:
            fname: file name relative to simulation case root directory.

        Returns:
            Dictionary with settings in the specified file.
        """
        return foam.read_dict(foam.getfilename(path.join(self._path, fname)))

    def write_data(
        self,
        fname: str,
        data: Iterable[Iterable[Any]],
        header: Iterable[Any] = "",
        fmode: str = "w",
    ) -> NoReturn:
        """Wrapper for write_data function in foam module.

        Args:
            data: the data to be written to file
            fname: the file name of the output data file
            fmode: the mode of the output file
            header: field names to be written as header

        Returns:
            Nothing is returned.
        """
        foam.write_data(fname, data, header, fmode)

    def exec(
        self,
        cmd: str,
        args: list[str],
        suffix: str = "",
        append: bool = False,
        overwrite: bool = False,
    ) -> bool:
        """Execute a command on case.

        A command is executed on the case and will behave similar to
        runApplication from RunFunctions of the OpenFOAM Foundation Software.
        Depending on the success of the command execution a 'End' or
        'FOAM FATAL ERROR' is added to the end of the log file.

        Args:
            cmd: command to be executed
            args: additional arguments for command
            suffix: suffix for log file
            append: append output to existing log file
            overwrite: overwrite existing log file

        Returns:
            True if command succeeds.
        """
        # Create log file name
        fmode = "w"
        file_log = path.join(
            self._path,
            ".".join(["log", path.basename(cmd), suffix]).rstrip("."),
        )
        log.info(f"Command output will be written to {file_log}")

        # Check for existing log file
        if append and not overwrite:
            log.debug("Appending to an existing log file.")
            fmode = "a"
        elif overwrite and not append:
            log.debug("Overwriting an existing log file.")
        elif overwrite and append:
            log.error("Overwrite and append option exclude each other.")
            return False
        else:
            if path.exists(file_log):
                log.error("Remove existing log file to re-run.")
                return False

        with open(file_log, fmode, encoding="utf-8") as f:
            try:
                log.info(f"Running '{cmd} {' '.join(args)}'.")
                run([cmd] + args, check=True, stdout=f, stderr=f)
                f.writelines(["\nEnd\n"])
            except CalledProcessError as e:
                log.debug(e)
                log.error(f"Writing error to {file_log}")
                f.writelines(["\nFOAM FATAL ERROR\n"])
                return False
            except FileNotFoundError:
                log.error(f"Can not find executable {cmd}.")
                return False

        return True


class FoamExtCase(FoamBaseCase):
    """An extended simulation case.

    Class contains parameters and functions for an extended simulation case.
    """

    def __init__(
        self, dname: str, mfile: str = "case.yml", mvalues: dict = None
    ) -> NoReturn:
        """Initialize extended OpenFOAM case.

        Initializes an extended OpenFOAM case, check for validity of the case
        and store some basic parameters. The metadata file stores some
        additional parameters for post-processing. The metadata file can use the
        Jinja2 template engine.

        Args:
            dname: directory containing the extended OpenFOAM case.
            mfile: YAML file containing metadata for case. In case the file is
                   not found, check if a file with extension .orig exists, if
                   not fail with error.
            mvalues: values to render templated YAML file with Jinja2.

        Parameters:
            _mfile (str): YAML file containing case information.
            _dump_mfile (str): file name, for the case that a .orig file
                               extension is used, to store processed YAML file.
            _metadata (Dict[str, str]): Metadata extracted from YAML file.


        Returns:
            Nothing.
        """
        super().__init__(dname)

        self._dump_mfile = mfile
        self._mfile = path.basename(foam.getfilename(path.join(dname, mfile)))

        if mvalues is None:
            self._mvalues = {}
        else:
            self._mvalues = mvalues
        self._metadata = {}

        # Read metadata file and render through Jinja2
        log.info(f"Reading metadata from {path.join(self._path, self._mfile)}.")
        self._metadata = self._mread(self._path, self._mfile, self._mvalues)

        # Store processed metadata file if required
        if self._dump_mfile != self._mfile:
            fpath = path.join(self._path, self._dump_mfile)
            if path.isfile(fpath):
                log.warning(
                    f"Processed metadata cannot be dumped to \
                            {fpath} because file exists."
                )
            else:
                log.info(f"Write processed metadata to {fpath}. ")
                with open(fpath, "w") as f:
                    yaml.dump(self._metadata, f)

        # Reduce metadata file to case specific section
        try:
            self._metadata = self._metadata[self._name]
        except KeyError:
            log.error(f"There is no key for simulation case: {self._name}.")
            exit(1)

        # Replace name if required
        if "name" in self._metadata:
            self._name = self._metadata["name"]

    def _mread(self, fpath: str, fname: str, fvalues: dict) -> Dict[str, str]:
        """Reading YAML file with metadata.

        Function to read YAML metadata from a given file.

        Args:
            fpath: directory containing the YAML file.
            fname: the YAML file name.
            fvalues: values to render templated YAML file with Jinja2.

        Returns:
            Returns a dictionary with the metadata.
        """
        env = Environment(loader=FileSystemLoader(fpath))
        template = env.get_template(fname)
        metadata = yaml.safe_load(template.render(fvalues))

        if not metadata:
            log.error("No valid metadata has been found.")
            exit(1)

        return metadata

    def keywords(self) -> Dict[str, str]:
        """Returning keyword list.

        Function to return the keywords from metadata.

        Returns:
            Returns a list with keywords.
        """
        try:
            return self._metadata["keywords"]
        except KeyError:
            log.error(f"No keywords defined in {self._mfile}.")
            exit(1)

    def checkkeywords(self, reference: Dict) -> bool:
        """Check keywords against reference.

        Args:
            reference: dictionary with reference keywords

        Returns:
            True if all keywords are in reference list.
        """
        try:
            kwords = self.keywords()
            log.debug(f"Case keywords: {kwords}")
            for category in kwords.keys():
                current = set(kwords[category])
                other = set(reference[category])
                diff = current.difference(other)
                if diff:
                    log.error(f"Unknown keyword(s): {', '.join(diff)}.")
                    return False
        except KeyError:
            log.error(f"Unknown category: {category}.")
            return False

        return True

    def validation_fields(self) -> List[str]:
        """Returning validation field list.

        Function to return the validation fields from metadata.

        Returns:
            Returns a list with the validation field names.
        """
        return sorted(self._metadata["validation"])

    def validation_file(self, field: str, key: str) -> str:
        """Name of the data file for a field and key.

        Function to return the file for a given field from key in metadata.

        Args:
            field: field name for validation data file.
            key: key for data file.

        Returns:
            Returns a file name.
        """
        return self._metadata["validation"][field][key]

    def validation_data(
        self, field: str, key: str
    ) -> Tuple[ndarray, Sequence[str]]:
        """Data for a field and key.

        Function to return the data included in the simulation file for a given
        field.

        Args:
            field: field name for validation data file.
            key: key for data file.

        Returns:
            A tuple of column names and a numpy data array with validation
            data.
        """
        filepath = path.join(self._path, self.validation_file(field, key))
        return foam.read_data(filepath)
