# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Create a customizable developer environment based on docker images.

This module provides the functionality to create a customizable environment
for docker images by using Docker SDK for Python.
"""

from subprocess import (
    run,
    CalledProcessError,
    DEVNULL,
    STDOUT,
    TimeoutExpired,
    PIPE,
)
from sys import exit, argv
from docker import DockerClient, from_env
from docker.models.images import Image as DockerImage
from docker.errors import (
    DockerException,
    ImageNotFound,
    APIError,
    BuildError,
    NotFound,
    NullResource,
)
from rich.console import Console
from rich.progress import Progress
from rich.markdown import Markdown
from jsonargparse import ArgumentParser, ActionConfigFile, Namespace
from typing import List, NoReturn, Tuple, Dict
from io import BytesIO
from json import loads
from socket import gethostname
from time import sleep
from rich_argparse import RichHelpFormatter
from ._version import __version__
import os
import logging
import re
import shutil
import pathlib
import pwd
import grp

# Start a console for beautiful output
console = Console()

# Enable logging for this module
log = logging.getLogger(__name__)


def addenvvars(envvars: List[str]) -> List[str]:
    """Add environment variables to docker image.

    Takes the arguments from the command line and translates them into docker
    options to add environment variables. In case there are already environment
    variables in the argument, they are evaluated on the host system.

    Args:
        envvars: list of environment variables to be added (format: var=value)

    Returns:
        List of options for docker command
    """
    opts = []
    try:
        for env in envvars:
            name, value = env.split("=")
            value = os.path.expandvars(value)
            log.debug(f"Adding environment variable {name} with value {value}")
            opts.extend(["--env", f"{name}={value}"])
    except ValueError as e:
        log.warning(
            f"An environment variable should be 'name=value' and not '{env}'"
        )
        log.debug(e)

    return opts


def adddisplay() -> List[str]:
    """Add display support.

    Depending on the X-Server and the host platform, support for graphical user
    interfaces (GUI) is enabled. At the moment only Linux with X11 and XWayland,
    and MacOS with XQuartz are supported.

    Returns:
        List of options for docker command
    """
    opts = []
    sysname = osname()
    xauthpath = f"/tmp/.docker.xauth.{os.getpid()}"

    if sysname == "darwin":
        log.debug("Enabling display support for Mac OS.")
        if xquartz():
            display = f"{hostname()}:0"
            if xauth(display, xauthpath):
                opts.extend(
                    [
                        "--env",
                        f"DISPLAY={display}",
                        "--env",
                        f"XAUTHORITY={xauthpath}",
                        "--volume",
                        f"{xauthpath}:{xauthpath}",
                    ]
                )

    elif sysname == "linux":
        log.debug("Enabling display support for Linux.")
        try:
            stype = os.environ["XDG_SESSION_TYPE"]
            log.debug(f"Found session type {stype}")
            if stype == "x11" or stype == "wayland":
                display = os.environ["DISPLAY"]
                if xauth(display, xauthpath):
                    opts.extend(
                        [
                            "--device",
                            "/dev/dri:/dev/dri",
                            "--volume",
                            "/tmp/.X11-unix:/tmp/.X11-unix",
                            "--env",
                            f"DISPLAY={display}",
                            "--env",
                            f"XAUTHORITY={xauthpath}",
                            "--volume",
                            f"{xauthpath}:{xauthpath}",
                        ]
                    )
            elif stype == "tty":
                log.debug(
                    f"Display support disabled for session type: {stype}."
                )
            else:
                log.warning(
                    f"Unknown display manager {stype}. Display support \
                      disabled."
                )
        except KeyError as e:
            log.warning(
                "Cannot get required environment variables. Is XWayland \
                 running?"
            )
            log.debug(e)
    else:
        log.warning(
            f"Display support not possible for operating system: {sysname}."
        )

    return opts


def addgpu() -> List[str]:
    """Add support for Nvidia GPUs.

    If a Nvidia runtime environment is available, mount the graphic cards (GPU)
    to the container at runtime.

    Returns:
        List of options for docker command
    """
    opts = []

    if shutil.which("nvidia-smi") and shutil.which("nvidia-container-cli"):
        log.info("Adding support for NVIDIA graphic cards.")
        opts.extend(["--gpus", "all"])
    else:
        log.debug("No NVIDIA runtime environment found.")

    return opts


def addjupyter(port: str = "8888") -> List[str]:
    """Add support for Jupyter Server.

    If a jupyter server start is requested, add the required options. The
    options include a healthcheck to make sure the server is running and
    listening on the specified port.

    Args:
        port: port for the server to listen.

    Returns:
        List of options for docker command
    """
    log.info("Launching jupyter server in the background")
    log.info(f"Connection: http://127.0.0.1:{port}/lab")
    log.info("Password  : The name of a famous containerization software.")

    return [
        "-p",
        f"{port}:{port}",
        "--health-cmd",
        f"curl -f http://localhost:{port}/ || exit 1",
        "--health-interval",
        "5s",
    ]


def addssh() -> List[str]:
    """Add support for ssh connection.

    Add required variables and volumes to support ssh connections from inside
    the docker container.

    Returns:
        List of options for docker command
    """
    opts = []
    try:
        sshvar = os.environ["SSH_AUTH_SOCK"]
        sshdir = os.path.dirname(sshvar)
        opts += [
            "--env",
            f"SSH_AUTH_SOCK={sshvar}",
            "--volume",
            f"{sshdir}:{sshdir}",
        ]
    except KeyError:
        log.warning("SSH_AUTH_SOCK not set, disabling ssh support.")
    return opts


def addvolumes(bindvols: List[str], image: DockerImage) -> List[str]:
    """Bind volumes to docker image.

    Takes the arguments from the command line and translates them into docker
    options to bind volumes. In case there are environment variables in the
    definition of the source, they are evaluated on the host system. In case
    there are variables in the target, the docker image is inspected and the
    variables are evaluated based on this information.

    Args:
        bindvols: list of volumes to be binded (format: source:target)
        image: Docker image object, which can be inspected

    Returns:
        List of options for docker command
    """
    opts = []

    if not isinstance(bindvols, list):
        log.error(f"Specify a list for bindvols: {bindvols}.")
        return opts

    for v in bindvols:
        try:
            source, target = v.split(":")
            source = os.path.expandvars(source)
            target = expandimagevars(target, image)
            if not target.startswith("/"):
                log.warning(f"Target is not an absolute path: {target}.")
                target = f"/{target}"

            log.debug(f"Mounting volume from {source} to {target}.")
            opts.extend(["--volume", f"{source}:{target}"])
        except ValueError as e:
            log.error(f"Specify bind volume as 'source:target' and not '{v}'.")
            log.debug(e)

    return opts


def buildimage(
    c: DockerClient,
    fromimage: str,
    imagetag: str,
    recipe: List[str] = None,
    createdirs: List[str] = None,
    ccache: bool = False,
) -> DockerImage:
    """Build custom image based on docker build instructions.

    Build a new image based on the provided docker build instructions. The base
    image is automatically set to the given image and no extra FROM instruction
    is required. The custom image is also used to integrate the user management
    into the docker image. The host user, with the corresponding user id,
    and the primary group of the host user, with the corresponding id, are used
    to create a corresponding user in the docker image. Furthermore, to ensure
    write permissions for all mount points of the binded volumes, the
    corresponding directories are created.

    Args:
        c: docker client object.
        fromimage: image name to pull from registry; serve as from image.
        imagetag: name of the custom image to be build.
        recipe: additional build instructions for custom docker image.
        createdirs: directories to be create in custom image, e.g., mount points
        ccache: enable compiler cache

    Returns:
        True, if image was successfully build
    """
    uinfo = getuserinfo()
    uname = uinfo["username"]
    uid = uinfo["userid"]
    gid = uinfo["groupid"]
    gname = uinfo["groupname"]

    default_recipe = f"""\
FROM {fromimage}
USER "root"
# Set user specific environment variables
ENV HOME="/home/{uname}"
ENV WM_PROJECT_USER_DIR=\
"${{HOME}}/${{WM_PROJECT}}/{uname}-${{WM_PROJECT_VERSION}}"
ENV FOAM_RUN="${{WM_PROJECT_USER_DIR}}/run"
ENV FOAM_USER_APPBIN="${{WM_PROJECT_USER_DIR}}/platforms/${{WM_OPTIONS}}/bin"
ENV FOAM_USER_LIBBIN="${{WM_PROJECT_USER_DIR}}/platforms/${{WM_OPTIONS}}/lib"
ENV PATH="${{FOAM_USER_APPBIN}}:${{PATH}}"
ENV LD_LIBRARY_PATH="${{FOAM_USER_LIBBIN}}:${{LD_LIBRARY_PATH}}"
ENV XDG_RUNTIME_DIR="/run/user/{uid}"
# Create new user with sudo access
RUN apt-get update \\
    && apt-get install -qy --fix-missing --no-install-recommends sudo \\
    && rm -rf /var/lib/apt/lists/* \\
    && groupadd -f -g "{gid}" "{gname}" \\
    && useradd -l -u "{uid}" -g "{gname}" -m "{uname}" -s "/bin/bash" \\
    && echo "{uname}" ALL=\\(root\\) NOPASSWD:ALL > "/etc/sudoers.d/{uname}" \\
    && chmod 0440 "/etc/sudoers.d/{uname}" \\
    && touch "${{HOME}}/.sudo_as_admin_successful" \\
    && chown {uname}:{gname} ${{HOME}}/.sudo_as_admin_successful \\
    && mkdir -p "/run/user/{uid}" \\
    && chmod 0700 "/run/user/{uid}" \\
    && chown {uname}:{gname} /run/user/{uid} \\
    && if [ -d ${{HOME}}/.cache ]; then chown -R {uname}:{gname} \
"${{HOME}}/.cache"; fi
WORKDIR "${{HOME}}"
"""

    if createdirs:
        default_recipe += f"""\
# Create directories, e.g., for mountpoints
RUN mkdir -p { " ".join(createdirs) } \\
    && chown {uname}:{gname} { " ".join(createdirs) }\n"""

    if ccache:
        default_recipe += """\
# Create settings for compiler cache
RUN bash -c "echo -e \\"disable = false\\ncache_dir = \
${HOME}/.cache/ccache\\" > /etc/ccache.conf"\n"""

    default_recipe += f"USER {uname}\n"

    if recipe is None:
        dockerfile = default_recipe
    else:
        dockerfile = (
            default_recipe
            + "# User selected build instructions\n"
            + "\n".join(recipe)
        )

    log.debug(f"Using the following build instructions:\n{dockerfile}")

    try:
        f = BytesIO(dockerfile.encode("utf-8"))
        with console.status("Building image ..."):
            c.images.build(fileobj=f, rm=True, pull=False, tag=imagetag)
    except BuildError as e:
        log.debug(e)
        return False

    return True


def checkrestart(c: DockerClient, name: str) -> bool:
    """Check for a existing named container.

    Check if docker container with the given name exists, and has the status
    'running'. In case a container exists, but does not have the status
    'running', the container is force removed.

    Args:
        c: docker client object.
        name: docker container name.

    Returns:
        Information of running container exists.
    """
    try:
        container = c.containers.get(name)
        if container.attrs["State"]["Status"] == "running":
            log.debug(f"Found a running container {name}.")
            return True
        else:
            container.remove(force=True)
            log.debug(f"Removing stopped container {name}.")
            return False
    except (NotFound, NullResource):
        log.debug(f"Found no container {name}.")
        return False


def checkimagename(name: str) -> Tuple[str, str, str]:
    """Check for image registry, repository and tag.

    Split the given image url into image registry, image repository and image
    tag. Check if repository and tag meet the OCI standards defined in
    https://github.com/opencontainers/distribution-spec/blob/main/spec.md. In
    case no tag is given, use 'latest' as default.

    Args:
        name: docker image name

    Returns:
        Tuple of registry, repository and tag
    """
    oci_name = re.compile(
        r"[a-z0-9]+((\.|_|__|-+)[a-z0-9]+)*(\/[a-z0-9]+((\.|_|__|-+)[a-z0-9]+)*)*"
    )
    oci_tag = re.compile("[a-zA-Z0-9_][a-zA-Z0-9._-]+")

    # Strip tag
    try:
        s, tag = name.rsplit(":", maxsplit=1)
        if not oci_tag.fullmatch(tag):
            log.warning("Image tag does not match OCI standard.")
    except ValueError:
        s = name
        tag = "latest"
        log.debug("Found no tag in image name, assuming latest.")

    # Strip repository
    try:
        reg, repo = s.split("/", maxsplit=1)
        if not oci_name.fullmatch(repo):
            log.warning("Image repository does not match OCI standard.")
        if not (("." in reg) or (":" in reg)):
            raise ValueError
    except ValueError:
        log.warning("Cannot separate registry from repository in image name.")
        reg = None
        repo = s

    log.debug(f"Registry  : {reg}\nRepository: {repo}\nTag       : {tag}")

    return (reg, repo, tag)


def checklogin(c: DockerClient, registry: str) -> bool:
    """Check login to docker registry.

    Login via 'docker login' to docker registry. The function uses the docker
    login command directly because docker login via Docker Python SDK does
    not properly get the login credentials from the dockercfg_path variable, in
    particular for docker rootless installations.

    Args:
        c: docker client object.
        registry: docker registry to get the image from

    Returns:
        Return true if login was successful.
    """
    try:
        with console.status("Login to registry ..."):
            run(
                ["docker", "login", registry],
                stdout=DEVNULL,
                stderr=STDOUT,
                timeout=5,
                check=True,
            )
        log.info(f"Login to {registry} successful.")
    except CalledProcessError as e:
        log.error("Is the docker deamon running?")
        log.debug(e)
        return False
    except TimeoutExpired as e:
        log.error(
            f"Cannot login to {registry}. Please run 'docker login {registry}'."
        )
        log.debug(e)
        return False

    return True


def executeandexit(cmd: List[str]) -> NoReturn:
    """Run a command with subprocess and exit.

    A little wrapper to execute the given command via subprocess and exit the
    program with the return code from the given cmd.

    Args:
        cmd: command with options to be executed.
    """
    log.debug(cmd)
    try:
        exit(run(cmd).returncode)
    except Exception as e:
        log.error(e)
        exit(1)


def expandimagevars(path: str, image: str) -> str:
    """Expand environment variables in docker image.

    Expands the environment variables given by inspecting the docker image.
    Only variables that are defined in the docker image by ENV are available.
    A regular expression is used to detect the environment variables. The
    expression is inspired by the os.path.expandvars function.

    Args:
        path: path with variables
        image: image to be inspected

    Returns:
        String with expanded variables
    """
    regex = re.compile(r"\$(\w+|\{[^}]*\})", re.ASCII)
    vars = regex.findall(path)
    for var in vars:
        value = getimagevar(var.strip("{}"), image)
        if value:
            path = path.replace(f"${var}", value)
            log.debug(f"Replacing ${var} with {value}.")
        else:
            log.error(f"Variable {var} in docker image not defined.")

    return path


def getclient() -> DockerClient:
    """Establish communication with docker daemon.

    Detect a docker context on the system. Request the selected context
    from docker and create a DockerClient object for it. In case no context is
    found, try the recommended way and request information from environment.

    Returns:
        An object for a docker client or exit with error.
    """
    try:
        base_url = loads(
            run(
                ["docker", "context", "inspect"],
                check=True,
                capture_output=True,
            ).stdout
        )[0]["Endpoints"]["docker"]["Host"]
        log.debug(f"Connection to docker daemon with url: {base_url}.")
        client = DockerClient(base_url=base_url)
        return client
    except DockerException:
        pass
    except CalledProcessError as e:
        log.error("Is the docker deamon running?")
        log.debug(e)
        exit(1)

    try:
        client = from_env()
        log.info(
            "Connection to docker deamon configured from environment variables."
        )
        return client
    except DockerException as e:
        log.error("Connection to docker daemon not possible.")
        log.debug(e)
        exit(1)


def getimagevar(name: str, image: str) -> str:
    """Get environment variable from docker image.

    Inspect the docker image attributes to get the environment variables of
    the image and return the value of the environment variable.

    Args:
        name: name of the environemnt variable
        image: docker image to be inspected

    Returns:
        Value of environment variable
    """
    value = None
    for var in image.attrs["Config"]["Env"]:
        if var.startswith(f"{name}="):
            value = var.split("=")[1]
            break
    return value


def getuserinfo() -> Dict:
    """Get information about user executing the program.

    Request information about username, user id, group name, group id, and
    home directory of the current user from the operating system.

    Returns:
        Dictionary with all information, mentioned in description.
    """
    uid = os.getuid()
    gid = os.getgid()

    return {
        "username": pwd.getpwuid(uid).pw_name,
        "userid": uid,
        "groupname": grp.getgrgid(gid).gr_name,
        "groupid": gid,
        "home": pwd.getpwuid(uid).pw_dir,
    }


def getimage(
    c: DockerClient,
    name: str,
    pull: bool,
    build: bool,
    recipe: List[str],
    createdirs: List[str],
    ccache: bool,
) -> DockerImage:
    """Get docker image object.

    Search for the required docker image in registry, build a custom image
    if required and return an object for the docker image.

    Args:
        c: docker client object.
        name: image name including tag and registry
        pull: turn on/off to pull image from registry
        build: turn on/off to build custom image
        recipe: build instructions for custom docker image
        createdirs: create dirctories in custom image, e.g., mountpoints
        ccache: enable compiler cache

    Returns:
        Docker image object
    """
    suffix = "MPY"
    reg, repo, tag = checkimagename(name)

    # Disable pulling if no registry is given
    if reg is None:
        pull = False
        imagename = f"{repo}:{tag}"
        customname = f"{repo}:{tag}{suffix}"
    else:
        imagename = f"{reg}/{repo}:{tag}"
        customname = f"{reg}/{repo}:{tag}{suffix}"

    if pull and build:
        if not checklogin(c, reg):
            log.error(f"Login to {reg} failed.")
            exit(1)

        if pullimage(c, imagename):
            log.info(f"Pulling of image {imagename} was successfull.")
            try:
                c.images.remove(customname, True, False)
                log.debug("Removed existing custom image due to new pull.")
            except ImageNotFound:
                pass
        else:
            log.error(f"Pulling of image {imagename} has failed.")
            exit(1)
    else:
        log.debug("Pulling has been disable with command line option.")

    if build:
        if createdirs:
            createdirs = [d.split(":")[1] for d in createdirs]

        if buildimage(c, imagename, customname, recipe, createdirs, ccache):
            log.info(f"Building of image {customname} was successfull.")
        else:
            log.error(f"Build of custom image {customname} has failed.")
            exit(1)
    else:
        log.debug("Building has been disable with command line option.")

    try:
        image = c.images.get(customname)
    except ImageNotFound as e:
        log.error(f"Cannot find image {customname} in local registry.")
        log.debug(e)
        exit(1)

    return image


def hostname() -> str:
    """Get name of the host system in network.

    A little wrapper for socket.gethostname to allow mocking in unit tests.

    Returns:
        Hostname
    """
    return gethostname()


def osname() -> str:
    """Get name of operating system.

    A little wrapper for os.uname.sysname to allow mocking in unit tests.

    Returns:
        Name of the operating system, e.g., darwin or linux
    """
    return os.uname().sysname.lower()


def pullimage(c: DockerClient, name: str) -> bool:
    """Pull an image from registry.

    Pull an image from registry and show progress with a progress bar for
    downloading and extracting the layers. The progress bar is afterwards
    removed from the terminal.

    Args:
        c: docker client object.
        name: image to pull from registry

    Returns:
        True, if image was successfully pulled
    """
    pullstatus = True
    try:
        with Progress(transient=True) as progress:
            tasks = {}
            resp = c.api.pull(name, stream=True, decode=True)

            for line in resp:
                if line.get("error"):
                    log.error(line["error"])
                    pullstatus = False
                elif line["status"] == "Downloading":
                    id = line["id"]
                    total = line["progressDetail"]["total"]
                    completed = line["progressDetail"]["current"]
                    desc = f"[red][Download {id}"
                elif line["status"] == "Extracting":
                    id = line["id"]
                    total = line["progressDetail"]["total"]
                    completed = line["progressDetail"]["current"]
                    desc = f"[green][Extract  {id}"
                elif line["status"].startswith("Status: Image is up to date"):
                    log.debug(f"Image {name} is up to date.")
                    break
                else:
                    continue

                if id not in tasks.keys():
                    tasks[id] = progress.add_task(
                        desc, total=total, completed=completed
                    )
                else:
                    progress.update(
                        tasks[id],
                        completed=completed,
                        total=total,
                        description=desc,
                    )

            # Bugfix: rich.Progress will print a new line, if no task is added:
            #         https://github.com/Textualize/rich/issues/2762
            if not tasks:
                progress.add_task("", completed=100.0)
    except ImageNotFound as e:
        log.error(f"Image {name} not found in registry.")
        log.debug(e)
        pullstatus = False
    except APIError as e:
        log.error(f"No connection to registry when pulling {name}.")
        log.debug(e)
        pullstatus = False

    return pullstatus


def xquartz() -> bool:
    """Check if XQuartz server on MacOS is running.

    Check if the server is running, in case not, try to launch it. A little
    wrapper for subprocess.run, which has issues to be mocked on linux system.

    Returns:
        True, if XQuartz is running
    """
    try:
        run(["pgrep", "-q", "Xquartz"], check=True)
        log.debug("Found running XQuartz server.")
    except CalledProcessError:
        try:
            log.info("Launching XQuartz server.")
            run(
                ["open", "-j", "-a", "XQuartz"],
                check=True,
                stdout=DEVNULL,
                stderr=DEVNULL,
            )
            sleep(5)

        except CalledProcessError as e:
            log.warning("Cannot launch XQuartz.")
            log.debug(e)
            return False

    return True


def xauth(display: str, xauthpath: str) -> bool:
    """Extract magic cookie for authentication at X server.

    Get the magic cookie for the displayname from the Xauthority file. A
    potential hostname in the display variable has to be stripped off, otherwise
    xauth will not work correctly. For some reason xauth does not fail with an
    error code in case no valid magic cookie can be obtained, hence, we have
    to grep manually for an error message.

    Returns:
        True, if authentication was successful
    """
    if not os.path.isfile(xauthpath):
        pathlib.Path(xauthpath).touch()

    try:
        dpyname = f":{display.split(':')[1]}"
    except IndexError as e:
        log.warning(f"Wrong fromat for display variable: {display}.")
        log.debug(e)
        return False

    try:
        xauth = run(
            ["xauth", "nextract", "-", dpyname],
            stdout=PIPE,
            stderr=PIPE,
            check=True,
        )
        if (
            run(
                ["grep", "-q", "No matches found"], input=xauth.stderr
            ).returncode
            == 0
        ):
            raise CalledProcessError(1, "grep")
        run(
            ["xauth", "-f", xauthpath, "nmerge", "-"],
            input=xauth.stdout,
            check=True,
        )
    except CalledProcessError as e:
        log.warning("Creating authentication for access to X server failed.")
        log.debug(f"Run 'xauth list {dpyname}' for debugging.")
        log.debug(e)
        return False

    return True


def cmdargs(cmdargs: list) -> Namespace:
    """Process command line arguments.

    Process command line arguments. Mainly separated into a function for
    unit testing of reading of config files.

    Args:
        cmdargs: list with program name and command line arguments, e.g.,
                 sys.argv

    Returns:
        Namespace object with command line arguments
    """
    log.debug(f"Command line arguments with program name: {cmdargs}.")

    progname = os.path.basename(cmdargs.pop(0))

    description = """Launch a debian/ubuntu based docker container for a given
    image. The image will be pulled from a registry if it does not exist locally
    or is newer. Afterwards a new image is build to enable the user management
    in the docker container. Custom build instructions are supported, mainly
    intended to install additional packages. In case a name is set for the
    container, the registry is checked if a running container exists, and a new
    interactive shell is returned in the running container."""

    epilog = rf"""\
# Restart and Attach

It is possible to configure {progname} for attaching to running containers. The
switch to enable this feature is the --name option. Pass a name to the docker
container and {progname} will check if it finds a docker container, which is
already running, and will open a new bash inside this container.

# Custom Docker Image

It is possible to specify a reduced set of docker build instructions to build a
custom image, e.g., RUN or ENV. The image will get a new tag with the suffix
'MPY'. In case a newer image is pulled from the registry, the custom image will
be build automatically. The custom image is also used to add the host user as a
new user to the docker image. The reason is to enable the dynamic code feature
of the simulation software, which prevents running as root in the docker image.
A new user and a new group will be created, with the corresponding user id and
group id of the host system user. The user will get sudo permissions.

# Variable Expansion

## Variable Expansion on the Host

All variables specified as command line parameters are expanded on the host
system. If variables are specified in the configuration file, variable expansion
on the host system is enabled for the source definition in --volume, and for the
--env option.

## Variable Expansion in the Image

Some options enable expansion by inspecting the docker image. This feature is
enabled for the target definition in --volume and for the --workdir option. When
specifing this options as command line arguments, the dollar sign has to be
escaped with a backslash, otherwise the variables are expanded on the host
system.

# Autocompletion

It is possible register {progname} for your shell's completion framework by
running: `eval "$(register-python-argcomplete {progname})"`

# Jupyter Server

With the --jupyter option it is possible to launch a container with a jupyter
lab server listening on port 8888. To access the jupyter server use Visual
Studio Code or a web browser with the following url: http://127.0.0.1:8888/lab.
If the server is running remotely, use the ip address or hostname of the remote
machine instead: http://<ip-address>:8888/lab. The container will be started as
'detached' in the background. Use `docker container ls` and
`docker container stop` to stop the container. For security reasons the jupyter
lab server requires a password. The password is the name of a famous
containerization software in lower case letters.

# Examples

- Mount a directory as `run` directory for simulation software:
  `{progname} --volume+ /my/local/path:\${{WM_PROJECT_USER_DIR}}/run`
- Build a custom image, by adding x11-apps package to docker image for testing
  X11 capabilities: `{progname} --custom-build "['RUN sudo apt-get update',
  'RUN sudo apt-get install -qy x11-apps']" <image>`
- Run a container with jupyter lab server:
  `{progname} --jupyter --name server <image>`
- Stop a container with jupyter lab server: `docker container stop server`

# Bugs and Issues

https://codebase.helmholtz.cloud/fwdc/python/-/issues"""

    parser = ArgumentParser(
        prog=progname,
        description=Markdown(description, style="argparse.text"),
        epilog=Markdown(epilog, style="argparse.text"),
        formatter_class=RichHelpFormatter,
    )

    # Add logging option to command line args
    parser.add_argument(
        "--log",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        help="""\
Specify log level. The level of information is increasing as follows
ERROR, WARNING, INFO, DEBUG""",
        default="WARNING",
    )

    # Add positional arguments
    parser.add_argument(
        "image",
        help="""\
Name of the docker image, e.g., group/project/image:tag. In case no tag is
specified, the image with the tag 'latest' will be pulled.""",
    )

    # Add optional arguments
    parser.add_argument(
        "--no-gpu",
        action="store_false",
        dest="gpu",
        help="""\
Turn off automatic GPU and display support. Display support works on Linux
systems with X11 and XWayland. On Mac OS enable in settings for XQuartz 'Allow
connections from network clients'""",
    )

    parser.add_argument(
        "--no-ssh",
        action="store_false",
        dest="ssh",
        help="""Turn off automatic SSH support.""",
    )

    parser.add_argument(
        "--no-home",
        action="store_false",
        dest="home",
        help="""Turn off automatic mounting of HOME directory.""",
    )

    parser.add_argument(
        "--no-pull",
        action="store_false",
        dest="pull",
        help="""Do not pull image from remote registry.""",
    )

    parser.add_argument(
        "--no-build",
        action="store_false",
        dest="build",
        help="""\
Do not build image custom docker image. This will also set --no-pull.""",
    )

    parser.add_argument("--config", action=ActionConfigFile)

    parser.add_argument(
        "--name",
        help="""\
Give the container a custom name. Enables attach functionality in case a
running container with this name is found in the local registry.""",
    )

    parser.add_argument(
        "--custom-build",
        type=List[str],
        metavar="DOCKER INSTRUCTION(S)",
        help="""\
Specify docker build instructions to build a custom image. Do not specify FROM,
this will be done automatically. By default when specifying an argument value,
the previous value is replaced. However, in some cases it might be decided to
append to the list instead of replacing. This can be achieved by adding + as
suffix to the argument key.""",
    )

    parser.add_argument(
        "--extra-args",
        type=List[str],
        metavar="DOCKER OPTION(S)",
        help="""\
Specify extra options for docker run. See docker run --help for more
information. By default when specifying an argument value, the previous value
is replaced. However, in some cases it might be decided to append to the list
instead of replacing. This can be achieved by adding + as suffix to the
argument key.""",
    )

    parser.add_argument(
        "--volume",
        type=List[str],
        help="""\
Volumes to be mounted to the docker container. Use the default docker syntax as
source:target. Variable expansion is enabled, see below for details.""",
        metavar="source:target",
    )

    parser.add_argument(
        "--env",
        type=List[str],
        help="""\
Environment variables to be set in the container environment. Variable expansion
is enabled, see below for details.""",
    )

    parser.add_argument(
        "--site",
        action="store_true",
        help="""\
Mount the current working directory to the site directory of the simulation
software in the docker image.""",
    )

    parser.add_argument(
        "--ccache",
        action="store_true",
        help="""\
Enable usage of Ccache, which is a compiler cache. It speeds up recompilation by
caching previous compilations and detecting when the same compilation is being
done again. The default directory to store the cache is
CCACHE_DIR=${HOME}/.cache/ccache.""",
    )

    parser.add_argument(
        "--workdir",
        help="""\
Working directory inside the container. Variable expansion is enabled, see below
for details.""",
    )

    parser.add_argument(
        "--jupyter",
        action="store_true",
        help="""\
Launch jupyter lab server within the container. The container will be launched
in the background. See below for details.""",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"multiphasepy version: {__version__}",
        help="Display version number for multiphasepy package.",
    )

    # Note: Process only arguments, not the progame name
    return parser.parse_args(cmdargs)


def script_docker() -> NoReturn:
    """Launch docker container."""
    try:
        # Process arguments
        args = cmdargs(argv)

        # Enable logging feature
        log.parent.setLevel(getattr(logging, args.log, None))

        # Confirm docker installation
        if shutil.which("docker"):
            log.debug("Docker installation is available.")
        else:
            log.error("No docker installation found.")
            exit(1)

        # Get docker client
        client = getclient()

        # Attach to a running container
        if args.name and checkrestart(client, args.name):
            log.info(f"Attaching to running container {args.name}.")
            cmd = [
                "docker",
                "exec",
                "--interactive",
                "--tty",
                args.name,
                "bash",
                "-l",
            ]
            executeandexit(cmd)
        # Run a new container
        else:
            log.info("Starting a new container.")
            # Get docker image
            image = getimage(
                client,
                args.image,
                args.pull,
                args.build,
                args.custom_build,
                args.volume,
                args.ccache,
            )

            docker_options = [
                "run",
                "--volume",
                "/etc/localtime:/etc/localtime:ro",
                "--init",
            ]

            if args.jupyter:
                docker_options += ["--detach"]
                docker_cmd = ["jupyter", "lab"]
            else:
                docker_options += ["--interactive", "--tty"]
                docker_cmd = ["bash", "-l"]

            # Adding options based on command line arguments
            if args.jupyter:
                docker_options += addjupyter()
            if args.name:
                docker_options += ["--name", args.name]
            if args.volume:
                docker_options += addvolumes(args.volume, image)
            if args.env:
                docker_options += addenvvars(args.env)
            if args.gpu:
                docker_options += addgpu()
                docker_options += adddisplay()
            if args.extra_args:
                docker_options += args.extra_args
            if args.site:
                docker_options += addvolumes(
                    [
                        f"{pathlib.Path.cwd()}:${{WM_PROJECT_SITE}}/${{WM_PROJECT_VERSION}}"
                    ],
                    image,
                )
            if args.ssh:
                docker_options += addssh()
            if args.home:
                docker_options += addvolumes(
                    [f"{getuserinfo()['home']}:${{HOME}}"], image
                )
            if args.workdir:
                docker_options += [
                    "--workdir",
                    expandimagevars(args.workdir, image),
                ]

            cmd = ["docker"] + docker_options + [image.tags[0]] + docker_cmd
            executeandexit(cmd)
    except KeyboardInterrupt:
        log.info("Keyboard interrupt. Ending program.")
        exit(0)
