# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for Fuzzy Logic Metrics.

This module contains functions calculate metrics for a data array for later use
in fuzzy logic module.
"""
import numpy as np
from sys import exit
import logging

# Enable logging for this module
log = logging.getLogger(__name__)


def calc(
    mtype: str,
    simdata: np.ndarray,
    expdata: np.ndarray,
    interpolate: bool = True,
) -> np.float64:
    """Wrapper function to calculate given metric.

    This function calculates the metric specified as function parameter for the
    given data arrays for simulation and experimental data. The data arrays are
    two-dimensional arrays, with the coordinates as first dimension and the
    data points as second dimension.

    Args:
        mtype: the metric that should be calculated.
        simdata: two-dimensional array with simulation data.
        expdata: two-dimensional array with experimental data.
        interpolate: linear interpolation of simulation to experimental coords.

    Returns:
        error metric.
    """
    if interpolate:
        idata = np.interp(expdata[:, 0], simdata[:, 0], simdata[:, 1])
    else:
        idata = simdata[:, 1]

    if mtype == "mre":
        return mre(simdata=idata, expdata=expdata[:, 1])
    elif mtype == "pcc":
        return pcc(simdata=idata, expdata=expdata[:, 1])
    else:
        log.error(f"Unknown metric request: {mtype}")
        exit(1)


def mre(simdata: np.ndarray, expdata: np.ndarray) -> np.float64:
    """Calculate mean relative error.

    This function calculates the mean relative error (MRE) of two
    one-dimensional data arrays.

    Args:
        simdata: simulation data
        expdata: experimental data

    Returns:
        mean relative error.
    """
    return abs(simdata - expdata).mean() / expdata.mean()


def pcc(simdata: np.ndarray, expdata: np.ndarray) -> np.float64:
    """Calculate Pearson correlation coefficient.

    This function calculates the Pearson correlation coefficient (PCC) of
    two one-dimensional data arrays.

    Args:
        simdata: one-dimensional array with simulation data
        expdata: one-dimensional array with experimental data

    Returns:
        Pearson correlation coefficient.
    """
    if not np.any(simdata) or not np.any(expdata):
        log.warning(
            "Flat profile detected, returning zero for Pearson \
correlation coefficient."
            ""
        )
        return np.float64(0)
    else:
        return np.corrcoef(simdata, expdata)[0, 1]
