# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for Post-Processing of simulation results.

This module contains functions for post-processing of results produced by
the simulation software, e.g., time series and more.
"""
import numpy as np
import logging
from typing import Tuple, Sequence
from scipy.integrate import trapezoid

# Enable logging for this module
log = logging.getLogger(__name__)


def select_coordinates(
    header: Sequence[str], coords: Sequence[str]
) -> Tuple[Sequence[str], Sequence[str], np.ndarray]:
    """Gather the indices of matching coordinate names.

    If a list of coordinates is provided the matching entries from the header
    are collected.

    Args:
        header: header of the columns
        coords: sequence of candidates for coordinate names

    Returns:
        A tuple of a sequence of the data column names, the coordinate column
        names and a bool numpy array for the mask.
    """
    log.debug(f"Header: {header}.")
    log.debug(f"Candidate/s for coordinate/s: {coords}.")

    coordinates = [coord for coord in coords if coord in header]

    if len(coordinates) == 0:
        log.error("None of the coordinate candidate/s found in header.")
        return (None, None, None)

    log.debug(f"Matching coordinate column/s: {coordinates}.")

    # Filter for coordinate column, create mask and strip header
    mask = np.full(len(header), True, dtype=bool)
    i_s = []
    for coord in coordinates:
        i_s.append(header.index(coord))
    mask[i_s] = False
    result_header = [name for name in header if name not in coords]
    log.debug(f"Using columns {i_s} as coordinates.")

    return (result_header, coordinates, mask)


def average(
    data: np.ndarray, header: Sequence[str], base: Sequence[str]
) -> Tuple[np.ndarray, Sequence[str]]:
    """Calculate the average of a data series.

    This function calculates the average of a data series with the trapezoid
    rule. The function resturns 'None' and the original header in case there are
    either zero or multiple base coordinates found in the header.

    Args:
        data: data series data to be processed
        header: header of the columns in the data array
        base: sequence of candidates for the base coordinate for averaging

    Returns:
        A tuple of the numpy data array and a sequence of the column names.
    """
    avg_header, coords, mask = select_coordinates(header, base)

    if not coords:
        return (None, None)
    elif len(coords) > 1:
        log.error(
            f"Multiple base coordinates are defined: {coords}. Please, select a\
single base coordinate from the header. Use --help."
        )
        return (None, None)

    coord_mask = np.invert(mask)
    coord_range = np.ptp(data[:, coord_mask])

    return (
        trapezoid(data[:, mask], x=data[:, coord_mask], axis=0) / coord_range,
        avg_header,
    )


def gradient(
    data: np.ndarray, header: Sequence[str], coordinates: Sequence[str]
) -> Tuple[np.ndarray, Sequence[str]]:
    """Calculating the gradient of a data series.

    This function calculates the partial derivative for all combinations of
    value fields and coordinates. If no coordinates are provided the default
    coordinate names are selected, if present: 'Time', 'time', 'x', 'y', 'z',
    'distance'. The function resturns 'None' and the original header in case no
    matching coordinate name could be found.

    Args:
        data: data series data to be processed
        header: header of the columns in the data array
        coordinates: sequence of candidates for the coordinates to calculate the
            partial derivatives for

    Returns:
        A tuple of the numpy data array and a sequence of the column names.
    """
    grad_header, coords, mask = select_coordinates(header, coordinates)

    if not coords:
        return (None, None)

    xs_data = np.empty((np.shape(data)[0], 0))
    grads_data = np.empty((np.shape(data)[0], 0))

    result_header = coords.copy()

    for coord in coords:
        x_data = data[:, header.index(coord)]
        result_header.extend(
            ["d_" + name + "_d_" + coord for name in grad_header]
        )
        xs_data = np.column_stack((xs_data, x_data))
        grads_data = np.column_stack(
            (grads_data, np.gradient(data[:, mask], list(x_data), axis=0))
        )

    return np.column_stack((xs_data, grads_data)), result_header
