# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module to compare two repositories.

This module contains functions to compare two repositories. It generates a
file with lists for new files in the left repository, for files copied from the
right to the left repository and for files copied from the right repository but
modified in the left repository. A multiline C++ header as well as a multiline
bash header containing copyright statements is excluded from comparison.
"""

from filecmp import dircmp
from os import path, walk
from typing import BinaryIO, NoReturn, List
from sys import exit
import logging

# Enable logging for this module
log = logging.getLogger(__name__)


def rm_copyright(fptr: BinaryIO) -> NoReturn:
    """Remove header copyright.

    Strips a C++ header or a bash comment header from a given file pointer. The
    program exits if the header has no valid end statement.

    Args:
        fptr: File pointer to binary file.

    Returns:
        Nothing.
    """
    # Read first line
    line = fptr.readline()

    # C++ and OpenFOAM Foundation software dictionary header
    if line.startswith(bytes("/*", "utf-8")):
        while True:
            line = fptr.readline()
            if line.endswith(bytes("*/\n", "utf-8")):
                break
            if not line:
                log.error(
                    f"No valid copyright header found in file {fptr.name}"
                )
                exit(1)
    # Bash comment header
    elif line.startswith(bytes("#", "utf-8")):
        while True:
            line = fptr.readline()
            if not line.startswith(bytes("#", "utf-8")):
                break
            if not line:
                log.error(
                    f"No valid copyright header found in file {fptr.name}"
                )
                exit(1)
    # Other files
    else:
        pass


def do_cmp(fleft: str, fright: str) -> bool:
    """Compare two files.

    Compare the content of two files. Strip C++ or bash comment header before
    comparison. The function is a modified version of the do_cmp function in
    filecmp module.

    Args:
        fleft: Filename for left file.
        fright: Filename for right file

    Returns:
        True if files (except header) are identical, false otherwise.
    """
    bufsize = 8 * 1024
    with open(fleft, "rb") as fptr1, open(fright, "rb") as fptr2:
        rm_copyright(fptr1)
        rm_copyright(fptr2)
        while True:
            b1 = fptr1.read(bufsize)
            b2 = fptr2.read(bufsize)
            if b1 != b2:
                return False
            if not b1:
                return True


Dircmp = dircmp


def analyse(
    dcmp: Dircmp, new: List[str], copy: List[str], modified: List[str]
) -> NoReturn:
    """Compare directories and files in the given dircmp object.

    Recursive function to the directory stucture in the left and right
    repository. For each sub directory in the left repository the files are
    compared and analysed if they are new, modified or copied.

    Args:
        dcmp: Dircmp object containing the two directories to compare
        new: List of new files
        copy: List of copied files
        modified: List of modified files

    Returns:
        Returns nothing, but modifies the list for new, copy and modified.
    """
    for f in dcmp.left_only:
        lfile = path.join(dcmp.left, f)
        if path.isfile(lfile):
            new.append(lfile)
        elif path.isdir(lfile):
            for root, _, files in walk(lfile):
                for ff in files:
                    new.append(path.join(root, ff))
        else:
            pass

    for f in dcmp.common_files:
        lfile = path.join(dcmp.left, f)
        rfile = path.join(dcmp.right, f)
        if path.isfile(lfile):
            if do_cmp(lfile, rfile):
                copy.append(lfile)
            else:
                modified.append(lfile)

    for sub_dcmp in dcmp.subdirs.values():
        analyse(sub_dcmp, new, copy, modified)


def filereport(
    fname: str, new: List[str], copy: List[str], modified: List[str], root: str
) -> NoReturn:
    """Print comparison report to file.

    Print the content of the lists for new, copied and modified files to a
    markdown file. Makes all paths relative to root path and adds some
    markdown headlines.

    Args:
        fname: File name for markdown file
        new: List of new files
        copy: List of copied files
        modified: List of modified files
        root: root path for making file path relative

    Returns:
        Nothing.
    """
    # Sort entries in file lists
    new.sort()
    copy.sort()
    modified.sort()

    # Print file lists with relative path
    with open(fname, "w", encoding="utf-8") as f:
        print(
            "# File List for Multiphase Code Repository\n\n[[_TOC_]]\n", file=f
        )
        print("## New Files\n", file=f)
        print(*[path.relpath(i, start=root) for i in new], sep="\n", file=f)
        print("\n## Modified Files\n", file=f)
        print(
            *[path.relpath(i, start=root) for i in modified], sep="\n", file=f
        )
        print("\n## Copied Files with OpenFOAM Foundation Copyright\n", file=f)
        print(*[path.relpath(i, start=root) for i in copy], sep="\n", file=f)
