# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module to provide entry functions for scripts.

This module contains several functions, which can be called by scripts. The idea
is to have argparse only in this module and define the command line interface
for each script here.
"""
import json
import yaml
import argparse
import logging
import numpy as np
from os import path, getcwd
from typing import NoReturn
from rich_argparse import RichHelpFormatter
from .post import average, gradient
from .foam import convert, read_data, write_data
from .case import FoamExtCase, FoamBaseCase
from .fuzzy import evaluate, default_pcc, default_mre, default_goodness
from .rpcmp import dircmp, analyse, filereport
from .rodare import citationcff2rodare, codemeta2rodare, publish_version
from ._version import __version__

# Enable logging for this module
# Note: Using the root logger here, to pass configuration to other modules
log = logging.getLogger(__name__)


def general_argparse(**kwargs) -> argparse.ArgumentParser:
    """Parser object with --log and --version option.

    Args:
        kwargs: unspecified number of arguments to be passed through

    Returns:
        argparse object
    """
    parser = argparse.ArgumentParser(
        formatter_class=RichHelpFormatter, **kwargs
    )

    # Add logging
    parser.add_argument(
        "--log",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        help="""\
Specify log level. The level of information is increasing as follows ERROR,
WARNING, INFO, DEBUG""",
        default="WARNING",
    )
    # Add version
    parser.add_argument(
        "--version",
        action="version",
        version=f"multiphasepy version: {__version__}",
        help="Display version number for multiphasepy package.",
    )
    return parser


def case_argparse(**kwargs) -> argparse.ArgumentParser:
    """Parser object with --case option.

    Args:
        kwargs: unspecified number of arguments to be passed through

    Returns:
        argparse object
    """
    parser = general_argparse(**kwargs)

    parser.add_argument(
        "-c",
        "--case",
        default=getcwd(),
        help="folder with (templated) simulation case",
    )

    return parser


def extcase_argparse(**kwargs) -> argparse.ArgumentParser:
    """Parser object with --case and --value option.

    Args:
        kwargs: unspecified number of arguments to be passed through

    Returns:
        argparse object
    """
    parser = case_argparse(**kwargs)

    parser.add_argument(
        "-v",
        "--values",
        type=json.loads,
        default={},
        help="Values to evaluate templated file for case specific metadata \
        from stdin or json file",
    )

    return parser


def script_convert() -> NoReturn:
    """Convert data files between different formats.

    Allowed formats are dat and csv. For dat files the last comment line before
    the data block is interpreted as column header. The converter removes all
    comments from the file, except the column header.
    """
    parser = general_argparse(description=script_convert.__doc__)

    # Add positional arguments
    parser.add_argument(
        "input",
        type=str,
        help="input data file, e.g., input.dat",
    )
    parser.add_argument(
        "output",
        type=str,
        help="output data file, e.g., output.csv",
    )

    # Save arguments
    args = parser.parse_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))

    convert(args.input, args.output)


def script_fuzzy() -> NoReturn:
    """Evaluate goodness for simulation case."""
    parser = extcase_argparse(description=script_fuzzy.__doc__)

    # Add positional arguments
    parser.add_argument(
        "-pcc",
        type=json.loads,
        default=default_pcc,
        help="coefficients for pcc metric from stdin or json file",
    )
    parser.add_argument(
        "-mre",
        type=json.loads,
        default=default_mre,
        help="coefficients for mre metric from stdin or json file",
    )
    parser.add_argument(
        "-goodness",
        type=json.loads,
        default=default_goodness,
        help="coefficients for goodness metric from stdin or json file",
    )

    # Save arguments
    args = parser.parse_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))

    current = FoamExtCase(args.case, mvalues=args.values)

    fields = current.validation_fields()
    for field in fields:
        log.info(f"  Processing field {field}")
        result = evaluate(current, field, args.pcc, args.mre, args.goodness)

        # Print and write results
        resfile = path.join(
            current.path(), "postProcessing", parser.prog, field + ".csv"
        )
        current.write_data(resfile, [result.values()], result.keys())

        for key, value in result.items():
            log.info(f"    {key:20}: {value}")
        log.info(f"Results are written to: {path.relpath(resfile)}")


def script_rpcmp() -> NoReturn:
    """Compare two repositories.

    Generate a file with a lists for new files in the left repository, for files
    copied from the right to the left repository and for files copied from the
    right repository but modified in the left repository. A multiline C++ header
    as well as a multiline bash header containing copyright statements is
    excluded from comparison. The result is printed with respect to the left
    repository to a markdown file.
    """
    # Parse arguments
    parser = general_argparse(description=script_rpcmp.__doc__)

    # Add positional arguments
    parser.add_argument("left", type=str, help="Repository to be compared to")
    parser.add_argument("right", type=str, help="Repository to compare")
    parser.add_argument(
        "--output", type=str, help="output file", default="FILES.md"
    )

    # Save arguments
    args = parser.parse_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))

    new = []
    copy = []
    modified = []

    log.info(f"Comparing {args.left} with {args.right}.")
    dcmp = dircmp(args.left, args.right)
    analyse(dcmp, new, copy, modified)
    filereport(args.output, new, copy, modified, dcmp.left)

    log.info(f"Total number of files: {len(new) + len(copy) + len(modified)}")
    log.info(f"New files            : {len(new)}")
    log.info(f"Modified files       : {len(modified)}")
    log.info(f"Copied files         : {len(copy)}")


def script_kcheck():
    """Check keywords against a reference list."""
    # Parse arguments
    parser = extcase_argparse(description=script_kcheck.__doc__)

    # Add optional arguments
    parser.add_argument(
        "-ref", required=True, help="YAML file with reference keyword list"
    )

    # Save arguments
    args = parser.parse_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))

    case = FoamExtCase(args.case, mvalues=args.values)

    try:
        with open(path.join(getcwd(), args.ref)) as f:
            reference = yaml.safe_load(f)
        log.debug(f"Reference keywords: {reference}")
    except FileNotFoundError as e:
        log.error(f"Cannot find file for reference keywords: {args.ref}.")
        log.debug(e)
        exit(1)

    if case.checkkeywords(reference):
        log.info("All keywords are valid.")
    else:
        log.error("The keywords are not valid.")
        exit(1)


def script_shrun():
    """Wraps a given command.

    A command is wrapped similar to runApplication from RunFunctions of the
    OpenFOAM Foundation Software. The options and the behaviour correspond to
    the runApplication function. Depending on the success of the command
    execution an 'End' or 'FOAM FATAL ERROR' is added to the end of the log
    file.
    """
    # Parse arguments
    parser = case_argparse(
        description=script_shrun.__doc__,
        usage="%(prog)s [options] [--] cmd [-arg ...]",
    )

    # Add positional arguments
    parser.add_argument("cmd", help="Command to be wrapped.")

    # Add optional arguments
    parser.add_argument(
        "--append",
        action="store_true",
        help="Append output to an existing log file.",
    )

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite an existing log file.",
    )

    parser.add_argument(
        "--suffix",
        type=str,
        nargs=1,
        default=[""],
        help="Add a suffix to log file name.",
    )

    args, cmd_args = parser.parse_known_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))
    log.debug(f"Command to be executed: {args.cmd}")
    log.debug(f"Arguments for command: {cmd_args}")

    case = FoamBaseCase(args.case)
    case.exec(args.cmd, cmd_args, args.suffix[0], args.append, args.overwrite)


def script_post():
    """Run some math as post-processing on simulation data files."""
    # Parse arguments
    parser = general_argparse(description=script_post.__doc__)

    default_coordinates = ["Time", "time", "x", "y", "z", "distance"]

    subparsers = parser.add_subparsers(
        dest="func",
        title="function",
        description="Post-processing function to be executed on data.",
    )

    avgparser = subparsers.add_parser(
        "avg",
        help="Time average for time series using numpy.average.",
        formatter_class=RichHelpFormatter,
    )
    avgparser.add_argument("file", help="data file")
    avgparser.add_argument("--output", type=str, help="output file")
    avgparser.add_argument(
        "--base",
        type=str,
        action="append",
        help="Name of the base coordinate.",
    )

    gradparser = subparsers.add_parser(
        "grad",
        help="Gradient for data series using numpy.gradient.",
        formatter_class=RichHelpFormatter,
    )
    gradparser.add_argument("file", help="data file")
    gradparser.add_argument("--output", type=str, help="output file")
    gradparser.add_argument(
        "--coordinate",
        type=str,
        action="append",
        help="Names of the coordinate, parameter may be called multiple times.",
    )

    args = parser.parse_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))

    match args.func:
        case "avg":
            log.info(f"Calculating average for file {args.file}.")
            if not args.base:
                args.base = default_coordinates
            results, header = average(*(read_data(args.file)), args.base)

        case "grad":
            log.info(f"Calculating gradient for file {args.file}.")
            if not args.coordinate:
                args.coordinate = default_coordinates
            results, header = gradient(*(read_data(args.file)), args.coordinate)

        case None:
            log.error("Missing function parameter. Use --help.")
            exit(1)

    if results is None:
        exit(1)

    if args.output:
        fname = path.abspath(args.output)
    else:
        fname = f"{args.file}_{args.func}.csv"
    log.info(f"Writing results to {fname}.")

    if np.ndim(results) == 1:
        write_data(fname, [results], header)
    else:
        write_data(fname, results, header)


def script_publish():
    """Publish software in Rossendorf Data Repository (Rodare).

    An existing software publication in the Rossendorf Data Repository
    (Rodare, https://rodare.hzdr.de) will be updated. The metadata is obtained
    from either a codemeta.json or CITATION.cff file.
    """
    # Parse arguments
    parser = general_argparse(description=script_publish.__doc__)

    parser.add_argument("--id", required=True, help="Deposition id for Rodare")
    parser.add_argument(
        "--pubid", required=True, help="Publication id for Robis"
    )
    parser.add_argument("--url", required=True, help="URL for Rodare API")
    parser.add_argument("--token", required=True, help="Rodare access token")
    parser.add_argument("--meta", required=True, help="Metadata file")
    parser.add_argument(
        "--files", nargs="+", type=str, required=True, help="File(s) for upload"
    )

    # Add optional arguments
    parser.add_argument(
        "--pkgversion",
        type=str,
        help="Package version number (default: obtain from metadata file).",
    )

    args = parser.parse_args()

    # Enable logging feature
    log.parent.setLevel(getattr(logging, args.log, None))

    # Update metadata
    if path.isfile(args.meta):
        _, ftype = path.splitext(args.meta)

        if ftype == ".cff":
            with open(args.meta, "r", encoding="utf8") as f:
                codemeta_schema = yaml.safe_load(f)
            rodare_data = citationcff2rodare(codemeta_schema, args.pubid)
        elif ftype == ".json":
            with open(args.meta, "r", encoding="utf8") as read_file:
                codemeta_schema = json.load(read_file)
            rodare_data = codemeta2rodare(codemeta_schema, args.pubid)
        else:
            log.error(f"Unknown file format for metadata: {ftype}.")
            exit(1)
    else:
        log.error(f"File {args.meta} does not exist.")
        exit(1)

    # Overwrite version from command line argument
    if args.pkgversion:
        log.info(
            f"Using version number specified as argument: {args.pkgversion}"
        )
        rodare_data["metadata"]["version"] = args.pkgversion
    else:
        log.info(
            f'Using version number from {args.meta}: \
{rodare_data["metadata"]["version"]}'
        )

    publish_version(args.url, args.token, args.id, args.files, rodare_data)
