# This file is part of the Multiphase Python Repository by HZDR.
#
# Copyright (C) 2023 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
# Website: https://hzdr.de
#
# Multiphase Python Repository by HZDR is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Multiphase Python Repository by HZDR is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the Multiphase Python Repository by HZDR. If not, see
# <http://www.gnu.org/licenses/>.

"""Module for Fuzzy Logic.

This module contains classes and functions to evaluate the Fuzzy Logic.
The details of the fuzzy logic system are described in:

Hänsch, S. et al. (2021). A workflow for the sustainable development of
closure models for bubbly flows. Chemical Engineering Science 244, 116807.
"""

from typing import NoReturn, Dict
from collections import OrderedDict
import numpy as np
import skfuzzy as fuzz
import logging
from . import case, metrics

# Enable logging for this module
log = logging.getLogger(__name__)

default_pcc = {
    "high": [0.4, 1.0, 1.0],
    "medium": [-1.0, 0.4, 1.0],
    "low": [-1.0, -1.0, 0.4],
}

default_mre = {
    "low": [0.0, 0.0, 0.3],
    "medium": [0.0, 0.3, 1.0],
    "high": [0.3, 1.0, 1.0],
}

default_goodness = {
    "perfect": [0.7, 1.0, 1.0],
    "good": [0.5, 0.7, 0.9],
    "tolerable": [0.3, 0.5, 0.7],
    "bad": [0.1, 0.3, 0.5],
    "defect": [0.0, 0.0, 0.3],
}

default_validation_keys = ["simulation", "reference"]
default_metric_keys = ["mre", "pcc"]

FoamExtCase = case.FoamExtCase


def calcmetrics(fobj: FoamExtCase, field: str, mkeys: list) -> Dict:
    """Evaluate the metrics for fuzzy logic.

    Evaluates the metrics required for the fuzzy logic for a simulation case.

    Args:
        fobj: extended simulation case object.
        field: field to process fuzzy logic.
        mkeys: name(s) of the metrics to be calculated.

    Returns:
        Dictionary with metric values.
    """
    # Load experimental data
    expdata, expheader = fobj.validation_data(field, "experiment")
    expdata = expdata[:, [0, expheader.index(field)]]

    # Loop over validation keys
    d = {}
    for k in default_validation_keys:
        data, header = fobj.validation_data(field, k)
        data = data[:, [0, header.index(field)]]
        for m in mkeys:
            d[f"{k}-{m}"] = metrics.calc(m, simdata=data, expdata=expdata)

    return d


def calcgoodness(
    mvalues: Dict, mkeys: Dict, s_pcc: Dict, s_mre: Dict, s_goodness: Dict
) -> Dict:
    """Evaluate goodness based on fuzzy logic.

    Evaluates the goodness based on fuzzy logic for a given simulation case.

    Args:
        mvalues: values for the considered metrics.
        mkeys: name(s) of the metrics to be considered.
        s_pcc: membership function shape parameters for pcc metric.
        s_mre: membership function shape parameters for mre metric.
        s_goodness: membership function shape parameters for goodness.

    Returns:
        Dictionary with goodness values.
    """
    # Create metric objects and Fuzzy analyser
    analyzer = FuzzyAnalyzer(
        {
            "pcc": FuzzyMetric(-1.0, 1.0, 100, "tri", s_pcc),
            "mre": FuzzyMetric(0, 1.0, 100, "tri", s_mre),
        },
        FuzzyMetric(0, 1.0, 100, "tri", s_goodness),
        method="bisector",
    )

    d = {}
    for k in default_validation_keys:
        d[f"{k}-goodness"] = analyzer.analyze(
            {m: min(mvalues.get(f"{k}-{m}"), 1.0) for m in mkeys}
        )

    return d


def evaluate(
    fobj: FoamExtCase,
    field: str,
    s_pcc: Dict = default_pcc,
    s_mre: Dict = default_mre,
    s_goodness: Dict = default_goodness,
) -> Dict:
    """Evaluate fuzzy logic.

    Evaluates the fuzzy logic for a given OpenFOAM case.

    Args:
        fobj: OpenFOAM extended case object.
        field: field to process fuzzy logic.
        s_pcc: membership function shape parameters for pcc metric.
        s_mre: membership function shape parameters for mre metric.
        s_goodness: membership function shape parameters for goodness.

    Returns:
        Dictionary with results for metric and goodness values.
    """
    # Create empty dictionary
    d = OrderedDict()

    # Add name of simulation case
    d.update({"Name": fobj.name()})

    # Calculate metric values and add to dictionary
    d.update(calcmetrics(fobj, field, default_metric_keys))

    # Calculate goodness values and add to dictionary
    d.update(calcgoodness(d, default_metric_keys, s_pcc, s_mre, s_goodness))

    return d


class FuzzyMetric:
    """Abstract metric for Fuzzy Logic model.

    Class representing abstract metric for Fuzzy Logic model evaluation and to
    generate the membership functions.
    """

    def __init__(
        self, start: float, stop: float, res: int, m_type: str, m_sets: dict
    ) -> NoReturn:
        """Initialize FuzzyMetric.

        Args:
            start: start value for independent variable in membership function.
            stop: stop value for independent variable in membership function.
            res: resolution for independent variable in membership function.
            m_type: membership function type, e.g., triangle.
            m_sets: shape parameter for membership function generation.

        Parameters:
            _space (str): Independent variable for membership function.
            _categories (dict): Membership functions for different categories.

        Returns:
            Nothing.
        """
        self._space = np.linspace(start, stop, res)
        self._categories = {}
        for key in m_sets.keys():
            if m_type == "tri":
                self._categories[key] = fuzz.trimf(self._space, m_sets[key])
            # fuzz.trapmf(x2, [4, 6, 8, 10])
            # it makes sense to make _low_ trapezoidal

    def space(self) -> np.ndarray:
        """Independent variable.

        Independent variable for membership functions.

        Returns:
            Independent variable
        """
        return self._space

    def m_names(self) -> list:
        """Membership categories.

        Categories for membership functions.

        Returns:
            Returns fuzzy set names, like 'good', 'bad' and 'ugly'
        """
        return list(self._categories.keys())

    def m_values(self, key: str) -> np.ndarray:
        """Membership function values.

        Membership function values for specific category.

        Args:
            key: category for membership function

        Returns:
            Returns membership function values for certain set 'key' name
        """
        return self._categories[key]


class FuzzyAnalyzer:
    """Fuzzy Logic Controller.

    Class for setting up Fuzzy Logic controller. It operates with
    fuzzy sets <results> and real estimated crisp values for each metric
    for producing <target> crisp value as a result of defuzzification of
    the target metric.
    """

    def __init__(
        self, results: Dict, target: FuzzyMetric, method: str
    ) -> NoReturn:
        """Initialize FuzzyAnalyzer.

        Args:
            results: FuzzyMetric for information to be analyzed.
            target: FuzzyMetric for result of defuzzification.
            method: Method for the defuzzification step.

        Parameters:
            _method (str): Method for defuzzification.
            _results (Dict[FuzzyMetric]): FuzzyMetric for information to be
                                          analyzed.
            _target (FuzzyMetric): FuzzyMetric for result of defuzzification
            _impactvalue (dict): Degree of membership for a given crisp value
                                 in a certain fuzzy set.
            _activation (dict): Apply fuzzy rule to infer/activate output
                                membership functions.
            _points (list): Input set of crisp values.
            _aggregated (bool): Aggregated output membership function composed
                                of output membership functions for each fuzzy
                                rule via a max function.

        Returns:
            Nothing.
        """
        if method not in ["centroid", "bisector", "mom", "som", "lom"]:
            raise ValueError("Wrong defuzzification method")
        self._method = method
        self._results = results
        self._target = target
        self._impactvalue = {}
        self._activation = {}
        self._points = []
        self._aggregated = None

    def analyze(self, points: dict) -> float:
        """Process Fuzzy Logic.

        Launch analysing process for Fuzzy controller if correct set of points
        provided. Points mean input set of crisp values.

        Args:
            points (dict): Input set of crisp values.

        Returns:
            Returns crisp result for target metric. For baseline model this
            means a value between 0 and 1 representing 'goodness' of the model.
        """
        if points.keys() != self._results.keys():
            raise ValueError("Unexpected keys values for evaluation points")

        self._points = dict(points)

        for key in self._results:
            point = self._points[key]
            metric = self._results[key]
            for f_set in metric.m_names():
                self._impactvalue[f"{key}:{f_set}"] = fuzz.interp_membership(
                    metric.space(), metric.m_values(f_set), point
                )

        self._aggregated = self.apply_rules()

        # Defuzzification
        # options centroid, mom, som, lom
        return fuzz.defuzz(self._target.space(), self._aggregated, self._method)

    def apply_rules(self):
        """Apply Fuzzy Rules.

        Method containing hardcoded rules for baseline Fuzzy controller:

        Baseline selection five fuzzy logic rules:

        1) vlow_confidence <-> defect model
           If PCC:low AND MRE:high

        2) low_confidence <-> bad model
           If (PCC:low AND MRE:medium) OR (PCC:medium AND MRE:high)

        3) medium_confidence <-> tolerable
           If (PCC:medium AND MRE:medium)
           OR (PCC:high AND MRE:high)
           OR (PCC:low AND MRE:low)

        4) high_confidence <-> good model
           If (PCC:high AND MRE:medium) OR (PCC:medium AND MRE:low)

        5) vhigh_confidence <-> perfect model
           If PCC:high AND MRE:low

        Returns:
            Aggregated function composed of the output function
            for each fuzzy rule.
        """
        # The five fuzzy rules
        vlow_confidence = np.fmin(
            self._impactvalue["pcc:low"],
            self._impactvalue["mre:high"],
        )

        low_confidence = np.fmax(
            np.fmin(
                self._impactvalue["pcc:low"],
                self._impactvalue["mre:medium"],
            ),
            np.fmin(
                self._impactvalue["pcc:medium"],
                self._impactvalue["mre:high"],
            ),
        )

        medium_confidence = np.fmax(
            np.fmin(
                self._impactvalue["pcc:medium"],
                self._impactvalue["mre:medium"],
            ),
            np.fmax(
                np.fmin(
                    self._impactvalue["pcc:high"],
                    self._impactvalue["mre:high"],
                ),
                np.fmin(
                    self._impactvalue["pcc:low"],
                    self._impactvalue["mre:low"],
                ),
            ),
        )

        high_confidence = np.fmax(
            np.fmin(
                self._impactvalue["pcc:high"],
                self._impactvalue["mre:medium"],
            ),
            np.fmin(
                self._impactvalue["pcc:medium"],
                self._impactvalue["mre:low"],
            ),
        )

        vhigh_confidence = np.fmin(
            self._impactvalue["pcc:high"],
            self._impactvalue["mre:low"],
        )

        # Apply the rules to the target set.
        # This step represents the inference of the target/output functions
        # as a consequence of each fuzzy rule.
        self._activation["very_low"] = np.fmin(
            vlow_confidence, self._target.m_values("defect")
        )
        self._activation["low"] = np.fmin(
            low_confidence, self._target.m_values("bad")
        )
        self._activation["medium"] = np.fmin(
            medium_confidence, self._target.m_values("tolerable")
        )
        self._activation["high"] = np.fmin(
            high_confidence, self._target.m_values("good")
        )
        self._activation["very_high"] = np.fmin(
            vhigh_confidence, self._target.m_values("perfect")
        )

        # Aggregate the output functions via max function.
        return np.fmax(
            self._activation["very_low"],
            np.fmax(
                self._activation["low"],
                np.fmax(
                    self._activation["medium"],
                    np.fmax(
                        self._activation["high"], self._activation["very_high"]
                    ),
                ),
            ),
        )
